// Org object

#import <stdlib.h>
#import <math.h>
#import <string.h>
#import <ctype.h>

#import "RDebug.h"

#import "Model.h"
#import "World.h"
#import "Org.h"
#import "Agent.h"

extern id	xZone;

// for collecting data in format of fortran program
int ChoiceHistory[GCanHistorySz][GCanHistorySz];
int DMHistory[GCanHistorySz][GCanHistorySz];
int ProbHistory[GCanHistorySz][GCanHistorySz];

@implementation Org

/////////////////////////////////////////////////////////////
// addOpenProblem
// addOpenChoice
//
// solveProblem: remove from openProblems list, 
//    add to solvedProblems.  set solveTime.
//    NOTE: incr activityCnt only if did NOT change attached Choice.
//
// addAgent: update "decstr" (ie add to choice eligibleAgents)
//           ==> NOTE: Org should establish its own list of Agents

-addOpenProblem: (id) p {
	[self setStrsForProblem: p];
	[openProblems addLast: p];
	return self;
}
-solveProblem: (id) p {
	id			index, o;
   	index = [openProblems begin: xZone];
   	while ( (o = [index next]) ) {
		if ( [p getId] == [o getId] ) {
			o = [index remove];
			break;
		}
   	} [index dropFrom: xZone];
	[solvedProblems addLast: p];
	[p setsolveTime: [model getT]];
    if ( [p getattachedChoice] == [p getattachedChoicePrev] )
        [p setactivityCnt: [p getactivityCnt] + 1];
	return self;
}

-addOpenChoice: (id) ch {
	[ch setrequiredE: (double) 1.1];	// ** Like fortran codes (960415)
	[self setStrsForChoice: ch];
	[openChoices addLast: ch];
	return self;
}

-addAgent: (id) Ag {
	char	c;
	id	   	index, ch;

	c = toupper( [model getCParValueFor: "decstr"] );
   	index = [openChoices begin: xZone];
   	while ( (ch = [index next]) ) {
		if ( [self agent: Ag eligibleForChoice: ch] )
			[ch addEligibleAgent: Ag];
   	} [index dropFrom: xZone];

	return self;
}

/////////////////////////////////////////////////////////
// makeChoice
// remove from openChoices, add to madeChoices
// set decisionTime and choiceMethod.
//
// set choiceMethod:
// R - resolution: some problems got solved
// O - oversite:   choice *never* had problems attached (max == 0)
// F - flight:     problems attached fled, then made ch. (max>0,cur=0)
// B - both flight and resolution (fled || max > cur) & ( cur > 0)
//     (have to have explicit fled marked in case problems came back)

-makeChoice: (id) ch {
	unsigned int cur, mx;

	// it will be removed from openChoices by caller.
	[madeChoices addLast: ch];
	[ch setdecisionTime: [model getT]];

	cur = [ch getattachedProblemsCnt];
	mx  = [ch getattachedProblemsCntMax];

	if ( mx == 0 ) { 
		[ch setchoiceMethod: 'O'];
	}
	else {  // had some Problems attached some point...
		if ( cur == 0 ) {	// all problems fled
			[ch setchoiceMethod: 'F'];
		}
		else { 
			if ( ![ch getfled] ) {
				[ch setchoiceMethod: 'R'];
			}
			else {
				[ch setchoiceMethod: 'B'];	// some prbs fled, some solved
			}
		}
	}

	return self;
}

/////////////////////////////////////////////////////////////
// setStrsForChoice: setup access and decision strs for new choice.
// 1. agents must all be checked to see if should be eligible
// 2. any open problems must be checked to see if should be eligible
// NB: smaller importance number is MORE important.

-setStrsForChoice: (id) ch {
	char	c;
	id	   	index, o;

	c = toupper( [model getCParValueFor: "decstr"] );
   	index = [agents begin: xZone];
   	while ( (o = [index next]) ) {
		DMsg5(2,"---setStrsForChoice %u: decstr %c, agent %u rank %u, ch importance %u.",
			[ch getId], c, [o getId], [o getrank], [ch getimportance] );
		if ( [self agent: o eligibleForChoice: ch] ) {
			[ch addEligibleAgent: o];
			DMsg(2," * add as eligible agent.");
		}
		DMsg(2,"\n");
   	} [index dropFrom: xZone];

	c = toupper( [model getCParValueFor: "accstr"] );
   	index = [openProblems begin: xZone];
   	while ( (o = [index next]) ) {
		DMsg5(2,"---setStrsForChoice %u: accstr %c, problem %u import %u, ch importance %u.",
			[ch getId], c, [o getId], [o getimportance], [ch getimportance] );
		if ( [self problem: o eligibleForChoice: ch] ) {
			[ch addEligibleProblem: o];
			DMsg(2," * add as eligible problem.");
		}
		DMsg(2,"\n");
   	} [index dropFrom: xZone];
	
	return self;
}

//////////////////////////////////////////////////////////////////////
// setStrsForProblem
// Any open Choices must be checked to see if this problem eligible.

-setStrsForProblem: (id) prob {
	char	c;
	id	   	index, ch;

	c = [model getCParValueFor: "accstr"];
   	index = [openChoices begin: xZone];
   	while ( (ch = [index next]) ) {
		DMsg5(2,"---setStrsForProblem %u: accstr %c, import %u, ch %u importance %u.",
			[prob getId], c, [prob getimportance], [ch getId], [ch getimportance] );
		if ( [self problem: prob eligibleForChoice: ch] ) {
			[ch addEligibleProblem: prob];
			DMsg(2," * add as eligible problem.");
		}
		DMsg(2,"\n");
   	} [index dropFrom: xZone];

	return self;
}

/////////////////////////////////////////////////////////////////////
// Agent/Problem-eligibleForChoice
// These implement the access-structure (Problem-Choice) and 
// decision-structure (Agent-Choice) control.
// 
// *** NOTE: The Pr eligibileForChoice: Ch depends on the setup
// ********  in the paper, namely there are twice as many problems
//           as choices, and import changes based on that.

-(BOOL) agent: (id) Ag eligibleForChoice: (id) Ch {
	unsigned importCh = [Ch getimportance];
	unsigned rankAg = [Ag getrank];
	char	c = toupper( [model getCParValueFor: "decstr"] );
   	if (   c == 'U' ||
   		 ( c == 'H' && rankAg <= importCh ) ||
   		 ( c == 'S' && rankAg == importCh )  ) {
		return 1;
   	}
	return 0;
}

-(BOOL) problem: (id) Pr eligibleForChoice: (id) Ch {
	unsigned importCh = [Ch getimportance] * 2;
	unsigned importPr = [Pr getimportance];
	char	c = toupper( [model getCParValueFor: "accstr"] );
   	if (   c == 'U' ||
		 ( c == 'H' && importPr <= importCh ) ||
		 ( c == 'S' && ( importPr == importCh || importPr == importCh-1 ) )
			  ) {
		return 1;
	}
	return 0;
}


/////////////////////////////////////////////////////////////
// getLeastNeedyChoiceForProblem
//
// NB: currently just take FIRST if there are ties 
//     (I think that's what paper does...)
//
// NOTE: there seem to be some problems in the mdc program,
//       relative to what the paper says. 
//       See Notes-1.txt of 19-20 Feb 1996.
// The program algorithm seems to be:
// for each open choice
//    if ( prob attached to THIS or NO choice before)
//       minimize ( reqE(ch) - attachedE(ch) )
//    else ( prob attached to some OTHER choice before! )
//       minimize ( rewE(ch) + reqE(pr) - attachedE(ch) )
// The paper claimed to minimize the deficit NOT including
// contribution of this problem.

-getLeastNeedyChoiceForProblem: (id) prob {
	id	   			index, ch, probOldCh, pickedCh;
	id				leastEChoices = [List create: xZone];
	unsigned int	leastEChoicesCnt = 0;
	double			deficit;
	double			leastNeeded = 999999999;
	FILE			*RF = [model getreportFILE];
	// extern id		uniformRandomR;

	fprintf(RF,"PL: %u", [prob getId] );
   	pickedCh = nil;
	index = [openChoices begin: xZone];
	probOldCh = [prob getattachedChoice];
   	while ( (ch = [index next]) ) {
		if ( [self problem: prob eligibleForChoice: ch] ) {
			// *** The deficit calculation ***
			if ( !probOldCh || ch == probOldCh ) {
				deficit = [ch getrequiredE] - [ch getattachedE];
				fprintf(RF," (%u reqCh(%.2f) - attCh(%.2f) => %.2f)",
				   [ch getId],[ch getrequiredE],[ch getattachedE],deficit );
			}
			else { // attached to some OTHER choice before
				deficit = [ch getrequiredE] + [prob getrequiredE]
                          - [ch getattachedE];
				fprintf(RF," (%u reqCh(%.2f) - attCh(%.2f) + reqPr(%.2f) => %.2f)",
				    [ch getId],[ch getrequiredE],[ch getattachedE],
				  	[prob getrequiredE], deficit );
			}
			DMsg4(2,"   ---getLeastNeedyChoiceForProblem %u: ch %u def %f, least %f.\n", [prob getId], [ch getId], deficit, leastNeeded );	
			// now decide whether this choice is better
			if ( ApproxEq(deficit,leastNeeded) ) {// equal the best, 
				DMsg(2,"   -- deficit == leastNeeded ...tie...\n" );
				[leastEChoices addLast: ch];
				++leastEChoicesCnt;
			}
			else if ( deficit < leastNeeded ) {		// new best
				DMsg(2,"   -- deficit < leastNeeded ...new least...\n" );
				leastNeeded = deficit;
				[leastEChoices drop];			// <== need removeAll !!
				leastEChoices = [List create: xZone];
				[leastEChoices addLast: ch];
				leastEChoicesCnt = 1;
			}
		}
   	} [index dropFrom: xZone];

	if ( leastEChoicesCnt > 0 ) {
		DMsg2(2,"   ---getLeastNeedyChoiceForProblem %u, ch cnt %u.",
			   [prob getId], leastEChoicesCnt );	
		index = [leastEChoices begin: xZone];
		while ( (ch = [index next]) ) {
			if ( !pickedCh )
				pickedCh = ch;
			else if ( [ch getimportance] < [pickedCh getimportance] )
				pickedCh = ch;
   		} [index dropFrom: xZone];
		// pickedCh = [leastEChoices atOffset: 0];
	   	// pickedCh = [leastEChoices atOffset: 
		//		[uniformRandomR rMax: leastEChoicesCnt] ];
		DMsg1(2," picked choice %u.\n", [pickedCh getId] );
	}
	[leastEChoices drop];
	fprintf(RF,"\n");
	return pickedCh;
}

///////////////////////////////////////////////////////////////
// getLeastNeedyChoiceForAgent
//
// NB: currently just take Most Important if there ties 
//     (I think that's what paper does...)
//// NOTE: there seem to be some problems in the mdc program,
//       relative to what the paper says. 
//       See Notes-1.txt of 19-20 Feb 1996.
// The program algorithm seems to be:
// for each open choice
//    if ( agent attached to THIS or NO choice before)
//       minimize ( reqE(ch) - attachedE(ch) )
//    else ( agent attached to some OTHER choice before! )
//       minimize ( reqE(pr) - attachedE(ch) - getE(agent,t-1)*SC(t-1) )
// where SC(t-1) is solnCoeffs for step t-1. 
// The paper claimed to minimize the deficit NOT including
// contribution of this agent.

-getLeastNeedyChoiceForAgent: (id) agnt {
	id	   			index, ch, agOldCh, pickedCh;
	id				leastEChoices = [List create: xZone];
	unsigned int	leastEChoicesCnt = 0;
	unsigned int    lastT = [model getT];
	double			leastNeeded = 999999999;
	double			deficit;
	FILE			*RF = [model getreportFILE];
	// extern id		uniformRandomR;

	if ( lastT < 1 ) lastT = 1;	// correction for first step!
	pickedCh = nil;

	fprintf(RF,"AL: %u", [agnt getId] );
	agOldCh = [agnt getattachedChoice];
	index = [openChoices begin: xZone];
   	while ( (ch = [index next]) ) {
		if ( [self agent: agnt eligibleForChoice: ch] ) {
			deficit = [ch getrequiredE] - [ch getattachedE];
			fprintf(RF," (%u req(%.2f) - att(%.2f) ", 
					[ch getId], [ch getrequiredE], [ch getattachedE] );
			if ( agOldCh != ch && agOldCh != nil ) {
				deficit -= [agnt getE] * 
						   [self getsolnCoeffForT: lastT];
				fprintf(RF," [-e(%.2f)*sc(%.2f)]", 
						[agnt getE], [self getsolnCoeffForT: lastT] );
			}
			fprintf(RF," => def %.2f)", deficit );
			if ( ApproxEq(deficit,leastNeeded) ) {
				[leastEChoices addLast: ch];
				++leastEChoicesCnt;
			}
			else if ( deficit < leastNeeded ) {
				leastNeeded = deficit;
				[leastEChoices drop];			// <== need removeAll !!
				leastEChoices = [List create: xZone];
				[leastEChoices addLast: ch];
				leastEChoicesCnt = 1;
			}
		}
   	} [index dropFrom: xZone];

	if ( leastEChoicesCnt > 0 ) {
		DMsg2(2,"   ---getLeastNeedyChoiceOfForAgent %u, ch cnt %u.",
			   [agnt getId], leastEChoicesCnt );
		index = [leastEChoices begin: xZone];
		while ( (ch = [index next]) ) {
			if ( !pickedCh )
				pickedCh = ch;
			else if ( [ch getimportance] < [pickedCh getimportance] )
				pickedCh = ch;
   		} [index dropFrom: xZone];
		// pickedCh = [leastEChoices atOffset: 0];
	   	// pickedCh = [leastEChoices atOffset: 
		//		[uniformRandomR rMax: leastEChoicesCnt] ];
		DMsg1(2," picked choice %u.\n", [pickedCh getId] );
	}
	[leastEChoices drop];
	fprintf(RF,"\n");
	return pickedCh;
}

///////////////////////////////////////////////////////////////
// createEnd -- set up some empty lists.
// Also set up the solution coefficients from the run par.
// There should be one solnCoeff for each T = 1 .. stopT.
-createEnd: (id) aZone {
	char			*cp, *stripFloat();
	float			sc;
	unsigned int	scCnt = 0;

	openChoices = [List create: aZone];
	openProblems = [List create: aZone];
	madeChoices = [List create: aZone];
	solvedProblems = [List create: aZone];
	agents = [List create: aZone];

	// get the one extra solnCoeff so can use use T=1..stopT.
    if ( !(solnCoeffs=(double *)calloc([model getstopT]+1, sizeof(double))) ) {
        SEMsg("\n\nERROR(Org-createEnd) calloc failed for solnCoeffs.\n\n" );
        exit( -1 );
    }
	cp = [model getParStringValueFor: "solncoeffs"];
	while ( (cp = stripFloat(cp,&sc)) && scCnt < [model getstopT]) {
		if ( sc <= 0.0 || sc > 1.0 ) {
			SEMsg1("\nERROR solnCoeff %.3f out of 0-1 range.\n", sc );
			exit( -1 );
		}
		++scCnt;
		solnCoeffs[scCnt] = sc;
	}
	if ( scCnt != [model getstopT] ) {
		SEMsg2("\nERROR: got %u solnCoeffs, not equal to stopT=%u of them!\n",
			   scCnt, [model getstopT] );
		exit( -1 );
	}

	return self;
}

//////////////////////////////////////////////////////////////////////////
// stepReport
// printFinalReport
//

-printstepReportFormat: (FILE *) RF {

	// NB: The CE and PE are printed as they enter, the
	//     next PL and AL are printed in getLeastNeedy*
	//     routines, so they are in this order in file.
	fprintf(RF,"# CE ChoiceEnter:    entryT Id Import\n" );
	fprintf(RF,"# PE ProblemEnter:   entryT Id Import\n" );
	fprintf(RF,"# PL ProblemLook:  ProbId (chId deficit) ...\n");
	fprintf(RF,"# AL AgentLook:   AgentId (chId deficit) ...\n");
	fprintf(RF,"# CM ChoiceMade:     entryT decT Id Import choiceMethod requiredE attachedE (AG-AgId Rank E)... (PR-ProbId Import requiredE)...\n" );
	fprintf(RF,"# PS ProblemSolved:  entryT solveT Id Import choiceMethod  ChoiceId\n" );
	fprintf(RF,"# CO ChoiceOpen:     entryT Id Import requiredE attachedE (AG-AgId Rank E)... (PR-ProbId Import requiredE)...\n" );
	fprintf(RF,"# PO ProblemOpen:    entryT Id Import AttachedCh\n" );

	return self;
}

-stepReport: (FILE *) RF {
	id			index, o, index1, o1;
	unsigned	T = [model getT];

	// choices made
	index = [madeChoices begin: xZone];
   	while ( (o = [index next]) ) {
		if ( [o getdecisionTime] == T ) {	// solved this step
			fprintf(RF,"CM: %u %u %u %u %c %.2f %.2f",[o getentryTime],T,[o getId],
				[o getimportance],[o getchoiceMethod],[o getrequiredE],[o getattachedE] );
			// list an attached agents...
			index1 = [[o getattachedAgents] begin: xZone];
			while ( (o1 = [index1 next]) ) {
		   		fprintf(RF," (AG-%u %u %.2f)", [o1 getId], 
							[o1 getrank], [o1 getE] );
				DMHistory[T][[o1 getId]] = [o getId];
			} [index1 dropFrom: xZone];
			// list any attached problems...
			index1 = [[o getattachedProblems] begin: xZone];
			while ( (o1 = [index1 next]) ) {
				fprintf(RF," (PR-%u %u %.2f)", [o1 getId], 
	   				[o1 getimportance], [o1 getrequiredE] );
				ProbHistory[T][[o1 getId]] = [o getId]; // not need this-mark below
			} [index1 dropFrom: xZone];
			fprintf(RF,"\n");
		}
		ChoiceHistory[T][[o getId]] = 2;
	} [index dropFrom: xZone];

	// problems solved
	index = [solvedProblems begin: xZone];
   	while ( (o = [index next]) ) {
		if ( [o getsolveTime] == T ) {	// solved this step
			fprintf(RF,"PS: %u %u %u %u %c %u\n", [o getentryTime], T, [o getId],
					[o getimportance], [[o getattachedChoice] getchoiceMethod],
					[[o getattachedChoice] getId] );
		}
		ProbHistory[T][[o getId]] = 1000;
	} [index dropFrom: xZone];

	// open problems
	index = [openProblems begin: xZone];
   	while ( (o = [index next]) ) {
		ProbHistory[T][[o getId]] = 0;
	} [index dropFrom: xZone];

	// Print open choices and attached problems
	index = [openChoices begin: xZone];
   	while ( (o = [index next]) ) {
	   	fprintf(RF,"CO: %u %u %u %.2f %.2f", [o getentryTime], [o getId],
		   		[o getimportance], [o getrequiredE], [o getattachedE] );
		ChoiceHistory[T][[o getId]] = 1;
	   	// list an attached agents...
		index1 = [[o getattachedAgents] begin: xZone];
		while ( (o1 = [index1 next]) ) {
	   		fprintf(RF," (AG-%u %u %.2f)", [o1 getId], 
	   				[o1 getrank], [o1 getE] );
			DMHistory[T][[o1 getId]] = [o getId];
		} [index1 dropFrom: xZone];
	   	// list any attached problems...
		index1 = [[o getattachedProblems] begin: xZone];
		while ( (o1 = [index1 next]) ) {
	   		fprintf(RF," (PR-%u %u %.2f)", [o1 getId], 
	   				[o1 getimportance], [o1 getrequiredE] );
			ProbHistory[T][[o1 getId]] = [o getId];
		} [index1 dropFrom: xZone];
		fprintf(RF,"\n");
	} [index dropFrom: xZone];

	return self;
}

-printFinalReport: (FILE *) RF {
	id			index, o;
	double		decCnt,decR,decO,decF,decB, agUnusedE, chWastedE;
	unsigned	chOpenCnt, chDiffCnt;
	unsigned	agActivCnt, agRestCnt;
	unsigned	prOpenCnt, prLatentCnt, prActivCnt, prAttachCnt;
	void WriteSpecialFortranOutput();
	
	decR = decO = decF = decB = decCnt = 0;
	agUnusedE = chWastedE = 0;
	chOpenCnt = chDiffCnt = 0;
	agActivCnt = agRestCnt = 0;
	prOpenCnt = prLatentCnt = prActivCnt = prAttachCnt = 0;

	// choices, open and made
	index = [openChoices begin: xZone];
   	while ( (o = [index next]) ) {
		++chOpenCnt;
		chDiffCnt += [model getT] - [o getentryTime] + 1;
	} [index dropFrom: xZone];

	index = [madeChoices begin: xZone];
   	while ( (o = [index next]) ) {
		decCnt += 1;
		if ( [o getchoiceMethod] == 'R' )
			decR += 1;
		else if ( [o getchoiceMethod] == 'O' )
			decO += 1;
		else if ( [o getchoiceMethod] == 'F' )
			decF += 1;
		else if ( [o getchoiceMethod] == 'B' )
			decB += 1;
		chWastedE += [o getattachedE] - [o getrequiredE];
		chDiffCnt += [o getdecisionTime] - [o getentryTime];
	} [index dropFrom: xZone];
	if ( decCnt > 0 ) {
		decR /= decCnt;
		decO /= decCnt;
		decF /= decCnt;
		decB /= decCnt;
	}

	// problems, open and solved
	index = [openProblems begin: xZone];
   	while ( (o = [index next]) ) {
		++prOpenCnt;
		prActivCnt += [o getactivityCnt];
		prLatentCnt += [o getlatencyCnt];
		prAttachCnt += [o getattachedPeriodCnt];

	} [index dropFrom: xZone];

	index = [solvedProblems begin: xZone];
   	while ( (o = [index next]) ) {
		prActivCnt += [o getactivityCnt];
		prLatentCnt += [o getlatencyCnt];
		prAttachCnt += [o getattachedPeriodCnt];
	} [index dropFrom: xZone];

	// agents
	index = [agents begin: xZone];
   	while ( (o = [index next]) ) {
		agActivCnt += [o getactivityCnt];
		agRestCnt += [o getrestingCnt];
		agUnusedE += [o getunusedE];
	} [index dropFrom: xZone];

	fprintf(RF,"#F  Org final report:\n");
	fprintf(RF,"#F  R=Resolution    O=Oversite     F=Flight        B=Both R,F\n" );
	fprintf(RF,"#F  KZ=chOpenCnt    KY=chDiffCnt   XS=chWastedE\n");
	fprintf(RF,"#F  KX=agActivCnt   KS=agRestCnt   XR=agUnsedE\n");
	fprintf(RF,"#F  KW=prOpenCnt    KV=prActivCnt  KU=prLatentCnt  KT=prAttachCnt\n");
	fprintf(RF,"#F     R    O    F    B   KZ  KY    XS   KX  KS    XR   KW  KV  KU  KT\n");
	fprintf(RF,"#FRN");
	fprintf(RF," %4.2f %4.2f %4.2f %4.2f", decR, decO, decF, decB );
	fprintf(RF," %3u %3u %6.2f", chOpenCnt, chDiffCnt, chWastedE );
	fprintf(RF," %3u %3u %6.2f", agActivCnt, agRestCnt, agUnusedE );
	fprintf(RF," %3u %3u %3u %3u",prOpenCnt,prActivCnt,prLatentCnt,prAttachCnt);
	fprintf(RF,"\n####\n");

	// Special output for comparing to fortran output
	WriteSpecialFortranOutput( model, chOpenCnt, chDiffCnt, agActivCnt, 
		prOpenCnt, prActivCnt, prLatentCnt, prAttachCnt, agRestCnt, agUnusedE, chWastedE );

	return self;
}


void WriteSpecialFortranOutput( id model, int KZ, int KY, int KX, int KW,
	int KV, int KU, int KT, int KS, double XR, double XS ) 
{
	id		o, index;
	int		t, n, DSint, ASint, ELoadint, EDistint;
	char	c, *cp, fileName[32];
	double  d;
    FILE	*fp;

    strncpy( fileName, [model getStrParValueFor: "inputParFileName"], 5 );
    fileName[5] = '\0';  strcat( fileName, "-fortran.out" );

    if ( strlen( fileName ) == 0 ) 
        strcpy( fileName, "gcan-fortran-format.out" );

    if ( !(fp = fopen( fileName, "a+" )) ) {
        SEMsg1("\n==> Can't open %s for appending fortran output!\n\n", fileName );
        return;
    }

	if ( (c = toupper([model getCParValueFor: "decstr"])) == 'U' )
		DSint = 0;
	else if ( c == 'H' ) DSint = 1;
	else DSint = 2;

	if ( (c = toupper([model getCParValueFor: "accstr"])) == 'U' )
		ASint = 0;
	else if ( c == 'H' ) ASint = 1;
	else ASint = 2;

	if ( (d=[model getDParValueFor: "eload"]) == 1.1 ) 
		ELoadint = 0;
	else if ( d == 2.2 ) ELoadint = 1;
	else ELoadint = 2;

	cp = [model getParStringValueFor: "agents"];
	if ( *cp == 'E' )
		EDistint = 1;
	else if ( *cp == 'H' ) EDistint = 2;
	else EDistint = 0;

	fprintf(fp, "1\n" );
	fprintf(fp, "  LD=%d PAcc=%d DStr=%d ED=%d   ", ELoadint, ASint, DSint, EDistint );
	fprintf(fp, " %d  %d  %d  %d  %d  %d  %d  %d  %.2f  %.2f", KZ, KY, KX, KW, KV,
				KU, KT, KS, XR, XS );
	fprintf(fp, "\n" );

	// choices open => choices made
	index = [[[model getorg] getmadeChoices] begin: xZone];
	n = 0;
   	while ( (o = [index next]) ) {
		if ( [o getchoiceMethod] == 'R' || [o getchoiceMethod] == 'B' )
			++n;
	} [index dropFrom: xZone];
	fprintf(fp, " IRNCR=%02d\n\n", n );

	fprintf(fp, "  CHOICE ACTIVATION HISTORY                 DEC.MAKER ACTIVITY HISTORY\n" );
	fprintf(fp, " 20 TIME PERIODS,10 CHOICES               20 TIME PERIODS,10 DEC.MAKERS\n");
	fprintf(fp, "  0=INACTIVE,1=ACTIVE,2=MADE                0=INACTIVE,x=WORKING ON CHOICE X\n\n");
	fprintf(fp, "           1 2 3 4 5 6 7 8 9 10                   1 2 3 4 5 6 7 8 9 10\n\n" );

	for ( t = 1; t <= 20; ++t ) {
		fprintf(fp, "     %2d   ", t );
		for ( n = 1; n <= 10; ++n )
			fprintf(fp,"%2u", ChoiceHistory[t][n] );
		fprintf(fp, "               %2d   ", t );
		for ( n = 1; n <= 10; ++n )
			fprintf(fp,"%2u", DMHistory[t][n] );
		fprintf(fp,"\n");
	}

	fprintf(fp,"\n PROBLEM HISTORY:ROWS=TIME,COLS=PROBS\n" );
	fprintf(fp," -1=NOT ENTERED,0=UNATTACHED,X=ATT.TO CH.X,**=SOLVED\n");
	fprintf(fp,"            1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20\n\n");
	for ( t = 1; t <= 20; ++t ) {
		fprintf(fp,"     %2d   ", t );
		for ( n = 1; n <= 20; ++n ) {
			if ( ProbHistory[t][n] < 1000 )
				fprintf(fp, " %2d", ProbHistory[t][n] );
			else
				fprintf(fp, " **" );
		}
		fprintf(fp, "\n" );
	}
	fprintf(fp, "\n" );
    if ( fp != stdout )
        fclose( fp );
}


/////////////////////////////////////////////////////////////////
// standard get/set methods

-(double) getsolnCoeffForT: (unsigned int) t {
		return solnCoeffs[t];
}

-(id) getopenChoices { return openChoices; }
-(id) getopenProblems { return openProblems; }
-(id) getmadeChoices { return madeChoices; }
-(id) getsolvedProblems { return solvedProblems; }
-(id) getagents { return agents; }
-(id) getmodel { return model; }
-(id) getworld { return world; }
-(id) getpop { return pop; }

-setagents: (id) o { agents = o; return self; }
-setmodel: (id) o { model = o; return self; }
-setworld: (id) o { world = o; return self; }
-setpop: (id) o { pop = o; return self; }

/////////////////////////////////////////////////////
// debugging aides

-displayOrgOn: (FILE *) RF {
	id			index, o, index1, o1;
	
	fprintf(RF,"openChoice  entryT  import   reqE  attE  needT\n" );
   	index = [openChoices begin: xZone];
   	while ( (o = [index next]) ) {
		fprintf(RF,"     %2u       %2u      %2u     %.2f  %.2f\n",
		   	[o getId], [o getentryTime], [o getimportance], [o getrequiredE],
			[o getattachedE] );
		fprintf(RF,"           EligAgnts: ");
		index1 = [[o geteligibleAgents] begin: xZone];
		while ( (o1 = [index1 next]) ) {
			fprintf(RF, " %u", [o1 getId] );
		}	[index1 dropFrom: xZone];
		fprintf(RF,"\n           EligProblems: ");
		index1 = [[o geteligibleProblems] begin: xZone];
		while ( (o1 = [index1 next]) ) {
			fprintf(RF, " %u", [o1 getId] );
		}	[index1 dropFrom: xZone];
		fprintf(RF,"\n           AttachedProblems: ");
		index1 = [[o getattachedProblems] begin: xZone];
		while ( (o1 = [index1 next]) ) {
			fprintf(RF, " %u", [o1 getId] );
		}	[index1 dropFrom: xZone];
		fprintf(RF,"\n           AttachedAgents: ");
		index1 = [[o getattachedAgents] begin: xZone];
		while ( (o1 = [index1 next]) ) {
			fprintf(RF, " %u", [o1 getId] );
		}	[index1 dropFrom: xZone];
		fprintf(RF,"\n");
	} [index dropFrom: xZone];

	fprintf(RF, "openProblems  entryT  import  reqE  actCnt latCnt\n" );
   	index = [openProblems begin: xZone];
   	while ( (o = [index next]) ) {
		fprintf(RF,"     %2u        %2u       %2u    %.2f    %2u      %2u\n",
		   	[o getId], [o getentryTime], [o getimportance], 
			[o getrequiredE], [o getactivityCnt], [o getlatencyCnt] );
	} [index dropFrom: xZone];

	fprintf(RF, "madeChoice  entryT  import   reqE  attE  decT diffclt Method\n" );
   	index = [madeChoices begin: xZone];
   	while ( (o = [index next]) ) {
		fprintf(RF,"     %2u       %2u      %2u     %.2f  %.2f    %2u",
		   	[o getId],[o getentryTime],[o getimportance],[o getrequiredE],
			[o getattachedE],[o getdecisionTime] );
		fprintf(RF,"    %u       %c\n", [o getdecisionTime] - [o getentryTime], 
                [o getchoiceMethod] );
		fprintf(RF,"           AttachedProblems: ");
		index1 = [[o getattachedProblems] begin: xZone];
		while ( (o1 = [index1 next]) ) {
			fprintf(RF, " %u", [o1 getId] );
		}	[index1 dropFrom: xZone];
		fprintf(RF,"\n           AttachedAgents: ");
		index1 = [[o getattachedAgents] begin: xZone];
		while ( (o1 = [index1 next]) ) {
			fprintf(RF, " %u", [o1 getId] );
		}	[index1 dropFrom: xZone];
		fprintf(RF,"\n");
	}

	fprintf(RF,"solvedProblems  entryT  import   reqE  actCnt latCnt  solveT\n" );
   	index = [solvedProblems begin: xZone];
   	while ( (o = [index next]) ) {
		fprintf(RF,"     %2u        %2u       %2u     %.2f    %2u     %2u      %2u\n",
		   	[o getId], [o getentryTime], [o getimportance], 
			[o getrequiredE], [o getactivityCnt], [o getlatencyCnt], 
            [o getsolveTime] );;
	} [index dropFrom: xZone];

	fprintf(RF,"Agents    E   actCnt  restCnt  unusedE\n" );
   	index = [agents begin: xZone];
   	while ( (o = [index next]) ) {
		fprintf(RF," %2u     %.2f    %2u      %2u      %.2f\n",
		   	[o getId], [o getE], [o getactivityCnt],
			[o getrestingCnt], [o getunusedE] ); 
	} [index dropFrom: xZone];

	return self;
}

@end

