// The World Object
// The Problem and Choice objects

#import <stdlib.h>
#import <math.h>
#import <string.h>

#import "RDebug.h"

#import "Model.h"
#import "Pop.h"
#import "Org.h"
#import "World.h"
#import "Agent.h"

extern id xZone;

@implementation World

//////////////////////////////////////////////////////////////
// activateNewProblemsFor, activateNewChoicesFor
// Check problems and choices list, and if any are to be
// activated this T, add them to lists in specified org.

-activateNewProblemsFor: (id) orgz ReportOn: (FILE *) RF {
   	id		index, p;
   	index = [problems begin: xZone];
   	while ( (p = [index next]) ) {
		if ( [model getT] == [p getentryTime] ) {
			[orgz addOpenProblem: p];
			fprintf(RF,"PE: %2u  %2u  %2u\n", [p getentryTime], [p getId],
					[p getimportance] );
		}
	} [index dropFrom: xZone];
	return self;
}

-activateNewChoicesFor: (id) orgz  ReportOn: (FILE *) RF {
   	id		index, ch;
   	index = [choices begin: xZone];
   	while ( (ch = [index next]) ) {
		if ( [model getT] == [ch getentryTime] ) {
			[orgz addOpenChoice: ch];
			fprintf(RF,"CE: %2u  %2u  %2u\n", [ch getentryTime], [ch getId],
			   	[ch getimportance] );
		}
	} [index dropFrom: xZone];
	return self;
}

////////////////////////////////////////////////////////////
// createEnd
// Set up the choices and problems.  The choices and  problems 
// parameters should consist of strings of values of the form:
// problems: entryTime,Importance, ...  (eg  1,8, 2,20, ...)
// choices:  entryTime,importance, ...
// Id's are same as importance if useii parameter is 1 (the default).
// Otherwise Id's are assigned in order from 1 as the objects are read.

-createEnd: (id) aZone {

	choices = [List create: aZone];
	problems = [List create: aZone];

	return self;
}

-initProblemsAndChoices: (id) aZone {
	char			*cp, *stripUI();
	unsigned int	nextId, et, import;
	id				o;

	DMsg(1,"-world-initPCs...\n");

	cp = [model getParStringValueFor: "choices"];
	nextId = 1;
	while ( (cp = stripUI(cp,&et)) ) {
		if ( (cp = stripUI(cp,&import)) ) {
			o = [Choice createBegin: aZone];
			if ( [model getBParValueFor: "useii"] )
				[o setId: import];
			else
				[o setId: nextId];
			[o setentryTime: et];
			[o setimportance: import];
			[o setorg: org];
			o = [o createEnd: aZone];
			[choices addLast: o];
			++nextId;
		}
		else {
			SEMsg("\nERROR (World-createEnd): parsing choices failed.\n");
			exit(-1);
		}
	}

	// now the problems
	cp = [model getParStringValueFor: "problems"];
	nextId = 1;
	while ( (cp = stripUI(cp,&et)) ) {
		if ( (cp = stripUI(cp,&import)) ) {
			o = [Problem createBegin: aZone];
			if ( [model getBParValueFor: "useii"] )
				[o setId: import];
			else
				[o setId: nextId];
			[o setentryTime: et];
			[o setimportance: import];
			[o setorg: org];
			o = [o createEnd: aZone];
			[problems addLast: o];
			++nextId;
		}
		else {
			SEMsg("\nERROR (World-createEnd): parsing problems failed.\n");
			exit(-1);
		}
	}

	if ( RDebug > 1 ) { // debugging
		id		index, o;
		char	buff[1024];
		SEMsg("--> Choices:\n" );
		index = [choices begin: aZone];
		while ( (o = [index next]) ) {
			[o displayInBuff: buff];
			SEMsg1("%s",buff);
		} [index dropFrom: aZone];
		SEMsg("--> Problems:\n" );
		index = [problems begin: aZone];
		while ( (o = [index next]) ) {
			[o displayInBuff: buff];
			SEMsg1("%s",buff);
		} [index dropFrom: aZone];
	}

	return self;
}

//////////////////////////////////////////////////////
// standard object access routines

-(id) getmodel { return model; }
-(id) getorg { return org; }
-(id) getpop { return pop; }

-setorg: (id) o { org = o; return self; }
-setmodel: (id) o { model = o; return self; }
-setpop: (id) o { pop = o; return self; }

/////////////////////////////////////////////////////
// debugging aides

@end

// ============================================================
// PROBLEM  PROBLEM PROBLEM  OBJECTS               PROBLEM OBJECTS 
// ============================================================
@implementation Problem 

/////////////////////////////////////////////////////////////
// tryAttachToOpenChoicesOfOrg
//
// Note: update attachedChoiceLast here, too.
//
// Note: don't unattach previous choice from this problem 
//   until we pick a  new choice for this problem  (or none).
//   Then and only if last choice TheProblem was attached to is
//   different from current choice should that
//   previous choice be detached now.

-tryAttachToOpenChoicesOfOrg: (id) Orgz {
	id	   	   	ch;

    attachedChoicePrev = attachedChoice;

	if ( (ch = [Orgz getLeastNeedyChoiceForProblem: self]) ) {
		DMsg2(3,"   ---tryAttachToOpenChoicesOfOrg prob %u, ch %u.\n",
			   Id, [ch getId] );
		if ( ch != attachedChoicePrev ) {
			attachedChoice = ch;
			[ch addAttachedProblem: self];
		}
	}

	// in any case, if this Problem was attached to a
	// *different* choice last step, tell it no longer attached.
   	if ( attachedChoicePrev && attachedChoicePrev != attachedChoice ) {
   		[attachedChoicePrev remAttachedProblem: self];
   	}
    // update this "problem changes" counter (to KV) (only from T=2 on)
    // note we also have to watch for case where problem enters
    // and is not attached to any choice!
    if ( [[org getmodel] getT] > 1 ) {
        if ( attachedChoicePrev != ch )
            ++activityCnt;
        if ( !attachedChoice && entryTime == [[org getmodel] getT] )
            ++activityCnt;
    }
	return self;
}


////////////////////////////////////////////////////////
// updateOpenStats - problem still open at end of T

-updateOpenStats {

	if ( attachedChoice ) {
		++attachedPeriodCnt;
	}
	else {
		++latencyCnt;
	}

	return self;
}

////////////////////////////////////////////////////////////
// createEnd

-createEnd: (id) aZone {

	attachedChoice = attachedChoicePrev = nil;
	requiredE = [[org getmodel] getDParValueFor: "eload"];
	activityCnt = latencyCnt = attachedPeriodCnt = 0;

	return self;
}

-setentryTime: (unsigned) t { entryTime = t; return self; }
-setimportance: (unsigned) i { importance = i; return self; }
-setId: (unsigned) i { Id = i; return self; }
-setsolveTime: (unsigned) t { solveTime = t; return self; }
-setrequiredE: (double) e { requiredE = e; return self; }
-setlatencyCnt: (unsigned int) c { latencyCnt = c; return self; }
-setorg: (id) o { org = o; return self; }
-setactivityCnt: (unsigned int) c { activityCnt = c; return self; }

-(unsigned int) getentryTime { return entryTime; }
-(unsigned int) getimportance { return importance; }
-(unsigned int) getId { return Id; }
-(unsigned int) getsolveTime { return solveTime; }
-(double) getrequiredE { return requiredE; }
-(unsigned int) getlatencyCnt { return latencyCnt; }
-(unsigned int) getactivityCnt { return activityCnt; }
-(unsigned int) getattachedPeriodCnt { return attachedPeriodCnt; }
-(id) getorg { return org; }
-(id) getattachedChoice { return attachedChoice; }
-(id) getattachedChoicePrev { return attachedChoicePrev; }

////////////////////////////////////////////////////////////
// 

-displayInBuff: (char *) buff {
	sprintf( buff, "Problem %2u:  importance %2u   entryTime %2u  solveTime %u  E required %.2f.\n",
			 Id, importance, entryTime, solveTime, requiredE );
	return self;
}

@end // ===========  end Problem object =======================

/////////////////////////////////////////////////////////////
// CHOICE  CHOICE   CHOICE  objct    CHOICE  CHOICE   CHOICE object
/////////////////////////////////////////////////////////////

@implementation Choice

-(unsigned int) getId { return Id; }
-(id) getorg { return org; }
-(unsigned int) getentryTime { return entryTime; }
-(unsigned int) getimportance { return importance; }
-(unsigned int) getdecisionTime { return decisionTime; }
-(double) getrequiredE { return requiredE; }
-(double) getattachedE { return attachedE; }
-(char) getchoiceMethod { return choiceMethod; }
-(id) geteligibleAgents { return eligibleAgents; }
-(id) getattachedAgents { return attachedAgents; }
-(id) geteligibleProblems { return eligibleProblems; }
-(id) getattachedProblems { return attachedProblems; }

-(unsigned int) getattachedProblemsCnt { 
	return attachedProblemsCnt; }
-(unsigned int) getattachedProblemsCntMax { 
	return attachedProblemsCntMax; }
-(BOOL) getfled { return fled; }

-setId: (unsigned) i { Id = i; return self; }
-setorg: (id) o { org = o; return self; }
-setentryTime: (unsigned) t { entryTime = t; return self; }
-setimportance: (unsigned) i { importance = i; return self; }
-setdecisionTime: (unsigned) t { decisionTime = t; return self; }
-setrequiredE: (double) e { requiredE = e; return self; }
-setattachedE: (double) e { attachedE = e; return self; }
-setfled: (BOOL) b { fled = b; return self; }
-setchoiceMethod: (char) c { choiceMethod = c; return self; }

-(double) getwastedE {
	if ( attachedE > requiredE )
		return attachedE - requiredE;
	else
		return (double) 0.0;
}

////////////////////////////////////////////////////////////
// createEnd

-createEnd: (id) aZone {

	attachedAgents = [List create: aZone];
	attachedProblems = [List create: aZone];
	eligibleAgents =  [List create: aZone];
	eligibleProblems =  [List create: aZone];
	fled = 0;
	attachedProblemsCnt = attachedProblemsCntMax = 0;

	return self;
}

//////////////////////////////////////////////////////////
// updateE of a choice
// see fortran 150-170
// requiredE = sum of requiredE of attached problems now
// attachedE = sum of E of attached agents now OR from the past!
//             NOTE: agent E is modified by solnCoeff for current T.

-updateE {
	id	   		index, a, p;

	requiredE = (double) 0.0;

	index = [attachedProblems begin: xZone];
   	while ( (p = [index next]) ) {
		requiredE += [p getrequiredE];
	} [index dropFrom: xZone];

	index = [attachedAgents begin: xZone];
   	while ( (a = [index next]) ) {
		attachedE += [a getE] * 
					 [org getsolnCoeffForT: [[org getmodel] getT]];
	} [index dropFrom: xZone];

	return self;
}


////////////////////////////////////////////////////////
// tryMakeChoice
// if neededE > 0, ignore.
// If neededE <= 0, then we can tell org to make this choice,
// and tell it to solve attached problems.
//
// In any case, update some stats on choice making!

-(BOOL) tryMakeChoice {
	BOOL			r = 0;
	unsigned int	pcnt = 0;
	id				index, p;
	double			neededE;

	DMsg1(2,"-tryMakeChoice %u...\n", Id );

	index = [attachedProblems begin: xZone];
   	while ( (p = [index next]) ) {
		++pcnt;
   	} [index dropFrom: xZone];
	if ( attachedProblemsCntMax > pcnt )
		fled = 1;
	if ( pcnt > attachedProblemsCntMax )
		attachedProblemsCntMax = pcnt;
	attachedProblemsCnt = pcnt;

	neededE = requiredE - attachedE; 
	if ( ApproxGE( (double) 0.0, neededE ) ) { // Make the choice!
		// before 90416: if ( neededE <= (double) 0.0 ) { // Make the choice!
		DMsg(3,"    -> send makeChoice to org!\n");
		[org makeChoice: self];		
		index = [attachedProblems begin: xZone];
		while ( (p = [index next]) ) {
			DMsg1(3,"   -> send solveProblem %u...\n", [p getId] );
			[org solveProblem: p];
		} [index dropFrom: xZone];
		DMsg1(3,"<= done making choice %u.\n", Id );
		r = 1;
	}

#ifdef THIS_NOT_WORK
	if ( r ) {
		[[org getopenChoices] remove: self];
	}
#endif

	return r;
}

/////////////////////////////////////////////////////////
//

-addAttachedAgent: (id) agnt {
	[attachedAgents addLast: agnt];
	return self;
}
-remAttachedAgent: (id) agnt {
	id	   		index, a;
	index = [attachedAgents begin: xZone];
   	while ( (a = [index next]) ) {
		if ( [a getId] == [agnt getId] ) {
			[attachedAgents remove: a];
			break;
		}
	} [index dropFrom: xZone];
	return self;
}
-addAttachedProblem: (id) prob {
	[attachedProblems addLast: prob];
	return self;
}
-remAttachedProblem: (id) prob {
	id	   		index, p;
	index = [attachedProblems begin: xZone];
   	while ( (p = [index next]) ) {
		if ( [p getId] == [prob getId] ) {
			[attachedProblems remove: p];
			break;
		}
	} [index dropFrom: xZone];
	return self;
}
-addEligibleAgent: (id) agnt {
	[eligibleAgents addLast: agnt];
	return self;
}
-addEligibleProblem: (id) prob {
	[eligibleProblems addLast: prob];
	return self;
}

////////////////////////////////////////////////////////////
// 

-displayInBuff: (char *) buff {
	sprintf( buff, "Choice %2u: import %2u  entryTime %2u  decTime %u  reqE %.2f  attE %.2f.\n",
			 Id, importance, entryTime, decisionTime, requiredE, attachedE );
	return self;
}

@end






