#!/usr/local/bin/tclsh
#
# This is a script to run a batch of GCan experimental runs.
# Try running with -man parameter for more info.

set prog      "/users/rlr/CAR/GCan/GCan.1/gcan"
set base      "x001-EL=1.1-AS=U";
set NRuns     1
set Debug     " -D=4 "

set eLoad { 1.1  }
set eDistr { E H S }
set decStr { U H S }
set accStr { U  }
set inputCases { case1.in-pars case2.in-pars case3.in-pars case4.in-pars }

set run			 0
set sleep		 0
set verbose		 0

#########################################################################
# Don't change things below here (unless you are re-doing this tool).
#########################################################################

set scriptName [file tail $argv0]
set DateTime [exec date +%C%m%d\-%H%M]

proc Usage args {
	global scriptName
	puts "Usage:  $scriptName \[-run\[=n\]\] \[-sleep\[=n\]\] \[-verbose\[=n\]\] \[others\]"
	puts "where:"
	puts "   -run       "
	puts "   -sleep     "
	puts "   -verbose   "
	puts "$scriptName -man     for manual type page."
    puts " "
}
##########################################################################
# Script documentation starts here.
#
##########################################################################

proc ManPage args {
	global scriptName

puts "
 No man page...
"
}

# first  process run time parameters for this batch script

for {set rtp  0} {$rtp < $argc} {incr rtp} {
	if { [string match "-run*" [lindex $argv $rtp]] } {
		set run 1
		if { [scan [lindex $argv $rtp] "-run=%d" runPar] == 1 } {
			set run $runPar
		}
	} elseif { [string match "-sleep*" [lindex $argv $rtp]] } {
		set sleep 2
		if { [scan [lindex $argv $rtp] "-sleep=%d" sleepPar] == 1 } {
			set sleep $sleepPar
		}
	} elseif { [string match "-verbose*" [lindex $argv $rtp]] } { 
		set verbose 3
		if { [scan [lindex $argv $rtp] "-verbose=%d" verbosePar] == 1 } {
			set verbose $verbosePar
		}
	} elseif { [string match "-man" [lindex $argv $rtp]] } {
		ManPage
		exit
	} else {
		puts "Unknown parameter '[lindex $argv $rtp]'."
		Usage
		puts "Bye."
		exit
	}
}


#
# Need confirmation messages here...
#
catch { exec mkdir $base }
catch { exec cp $scriptName $base }
foreach C $inputCases {
	catch { exec cp $C $base }
}
cd $base

# Now do the runs

foreach EL $eLoad {
	foreach ED $eDistr {
		foreach DS $decStr {
			foreach AS $accStr {
				foreach C $inputCases {

					set par	"-infile=$C -eload=$EL -agents=$ED -decstr=$DS -accstr=$AS $Debug"
					set parPrefix "EL=$EL-ED=$ED-DS=$DS-AC=$AS-$C"

					set logf	$parPrefix.lg
					set reportf	$parPrefix.report

					if { $run } {
						if { [catch { eval exec nice $prog $par -reportFileName=$reportf >& $logf } result ] } {
							puts stderr $result
						} elseif { $verbose > 1 } {
							puts "ran ok..."
						}
					} else {
						puts " eval exec nice $prog $par -reportFileName=$reportf > $logf"
					}
					if { $sleep } { exec sleep $sleep }
					if { $verbose > 1 } { puts "Parameter setting $par done." }
				}
			}
		}
	}
}

if { $verbose > 0 } { puts "Batch job Done." }



















