// getLeastNeedyChoiceForProblem
// NB: currently just take FIRST if there are ties 
//     (I think that's what paper does...)
// NOTE: there seem to be some problems in the mdc program,
//       relative to what the paper says. 
//       See Notes-1.txt of 19-20 Feb 1996.
// The program algorithm seems to be:
// for each open choice
//    if ( prob attached to THIS or NO choice before)
//       minimize ( reqE(ch) - attachedE(ch) )
//    else ( prob attached to some OTHER choice before! )
//       minimize ( rewE(ch) + reqE(pr) - attachedE(ch) )
// The paper claimed to minimize the deficit NOT including
// contribution of this problem.

-getLeastNeedyChoiceForProblem: (id) prob {
	id	   			index, ch, probOldCh;
	id				leastEChoices = [List create: xZone];
	unsigned int	leastEChoicesCnt = 0;
	double			deficit;
	double			leastNeeded = 999999999;

	index = [openChoices begin: xZone];
	probOldCh = [prob getattachedChoice];
   	while ( (ch = [index next]) ) {
		if ( [self problem: prob eligibleForChoice: ch] ) {
			// *** The deficit calculation ***
			if ( !probOldCh || ch == probOldCh ) {
				deficit = [ch getrequiredE] - [ch getattachedE];
			}
			else { // attached to some OTHER choice before
				deficit = [ch getrequiredE] + [prob getrequiredE]
                          - [ch getattachedE];
			}
			// now decide whether this choice is better
			if ( deficit < leastNeeded ) {		// new best
				leastNeeded = deficit;
				[leastEChoices drop];			// <== need removeAll !!
				leastEChoices = [List create: xZone];
				[leastEChoices addLast: ch];
				leastEChoicesCnt = 1;
			}
			else if ( deficit == leastNeeded ) {// equal the best
				[leastEChoices addLast: ch];
				++leastEChoicesCnt;
			}
		}
   	} [index dropFrom: xZone];

	if ( leastEChoicesCnt > 0 ) {
		DMsg2(2,"   ---getLeastNeedyChoiceForProblem %u, ch cnt %u.",
			   [prob getId], leastEChoicesCnt );
		ch = [leastEChoices atOffset: 0 ];
		// ch = [leastEChoices atOffset: 
		//		[uniformRandomR rMax: leastEChoicesCnt] ];
		DMsg1(2," picked choice %u.\n", [ch getId] );
	}
	[leastEChoices drop];

	return ch;
}
///////////////////////////////////////////////////////////////
// getLeastNeedyChoiceForAgent
// NB: currently just take FIRST if there are ties 
//     (I think that's what paper does...)
//// NOTE: there seem to be some problems in the mdc program,
//       relative to what the paper says. 
//       See Notes-1.txt of 19-20 Feb 1996.
// The program algorithm seems to be:
// for each open choice
//    if ( agent attached to THIS or NO choice before)
//       minimize ( reqE(ch) - attachedE(ch) )
//    else ( agent attached to some OTHER choice before! )
//       minimize ( reqE(pr) - attachedE(ch) - getE(agent,t-1)*SC(t-1) )
// where SC(t-1) is solnCoeffs for step t-1. 
// The paper claimed to minimize the deficit NOT including
// contribution of this agent.

-getLeastNeedyChoiceForAgent: (id) agnt {
	id	   			index, ch, agOldCh;
	id				leastEChoices = [List create: xZone];
	unsigned int	leastEChoicesCnt = 0;
	double			leastNeeded = 999999999;
	double			deficit;

	agOldCh = [agnt getattachedChoice];
	index = [openChoices begin: xZone];
   	while ( (ch = [index next]) ) {
		if ( [self agent: agnt eligibleForChoice: ch] ) {
			deficit = [ch getrequiredE] - [ch getattachedE];
			if ( agOldCh != ch ) {
				deficit -= [agnt getE] * 
						   [self getsolnCoeffForT: [model getT]];
			}
			if ( deficit < leastNeeded ) {
				leastNeeded = deficit;
				[leastEChoices drop];			// <== need removeAll !!
				leastEChoices = [List create: xZone];
				[leastEChoices addLast: ch];
				leastEChoicesCnt = 1;
			}
			else if ( deficit == leastNeeded ) {
				[leastEChoices addLast: ch];
				++leastEChoicesCnt;
			}
		}
   	} [index dropFrom: xZone];

	if ( leastEChoicesCnt > 0 ) {
		DMsg2(2,"   ---getLeastNeedyChoiceOfForAgent %u, ch cnt %u.",
			   [agnt getId], leastEChoicesCnt );
		ch = [leastEChoices atOffset: 0 ];
	   	// ch = [leastEChoices atOffset: 
		//		[uniformRandomR rMax: leastEChoicesCnt] ];
		DMsg1(2," picked choice %u.\n", [ch getId] );
	}
	[leastEChoices drop];

	return ch;
}
